"""Checks related to ansible specific best practices."""

import os
import re
from collections import defaultdict

from ansiblelater.command.candidates import Error
from ansiblelater.command.candidates import Result
from ansiblelater.command.candidates import Template
from ansiblelater.utils import count_spaces
from ansiblelater.utils.rulehelper import get_normalized_tasks
from ansiblelater.utils.rulehelper import get_normalized_yaml


def check_braces_spaces(candidate, settings):
    yamllines, errors = get_normalized_yaml(candidate, settings)
    conf = settings["ansible"]["double-braces"]
    description = "no suitable numbers of spaces (min: {min} max: {max})".format(
        min=conf["min-spaces-inside"], max=conf["max-spaces-inside"])

    matches = []
    braces = re.compile("{{(.*?)}}")

    if not errors:
        for i, line in yamllines:
            match = braces.findall(line)
            if match:
                for item in match:
                    matches.append((i, item))

        for i, line in matches:
            [leading, trailing] = count_spaces(line)
            sum_spaces = leading + trailing

            if (
                (sum_spaces < conf["min-spaces-inside"] * 2)
                or (sum_spaces > conf["min-spaces-inside"] * 2)
            ):
                errors.append(Error(i, description))
    return Result(candidate.path, errors)


def check_named_task(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    nameless_tasks = ["meta", "debug", "include_role", "import_role",
                      "include_tasks", "import_tasks", "include_vars",
                      "block"]
    description = "module '%s' used without or empty name attribute"

    if not errors:
        for task in tasks:
            module = task["action"]["__ansible_module__"]
            if ("name" not in task or not task["name"]) and module not in nameless_tasks:
                errors.append(Error(task["__line__"], description % module))

    return Result(candidate.path, errors)


def check_name_format(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    description = "name '%s' should start with uppercase"
    namelines = defaultdict(list)

    if not errors:
        for task in tasks:
            if "name" in task:
                namelines[task["name"]].append(task["__line__"])
        for (name, lines) in namelines.items():
            if name and not name[0].isupper():
                errors.append(Error(lines[-1], description % name))

    return Result(candidate.path, errors)


def check_unique_named_task(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    description = "name '%s' appears multiple times"

    namelines = defaultdict(list)

    if not errors:
        for task in tasks:
            if "name" in task:
                namelines[task["name"]].append(task["__line__"])
        for (name, lines) in namelines.items():
            if name and len(lines) > 1:
                errors.append(Error(lines[-1], description % name))

    return Result(candidate.path, errors)


def check_command_instead_of_module(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    commands = ["command", "shell", "raw"]
    modules = {
        "git": "git", "hg": "hg", "curl": "get_url or uri", "wget": "get_url or uri",
        "svn": "subversion", "service": "service", "mount": "mount",
        "rpm": "yum or rpm_key", "yum": "yum", "apt-get": "apt-get",
        "unzip": "unarchive", "tar": "unarchive", "chkconfig": "service",
        "rsync": "synchronize", "supervisorctl": "supervisorctl", "systemctl": "systemd",
        "sed": "template or lineinfile"
    }
    description = "%s command used in place of %s module"

    if not errors:
        for task in tasks:
            if task["action"]["__ansible_module__"] in commands:
                if "cmd" in task["action"]:
                    first_cmd_arg = task["action"]["cmd"].split()[0]
                else:
                    first_cmd_arg = task["action"]["__ansible_arguments__"][0]

                executable = os.path.basename(first_cmd_arg)
                if (first_cmd_arg and executable in modules
                        and task["action"].get("warn", True) and "register" not in task):
                    errors.append(
                        Error(task["__line__"], description % (executable, modules[executable])))

    return Result(candidate.path, errors)


def check_install_use_latest(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    package_managers = ["yum", "apt", "dnf", "homebrew", "pacman", "openbsd_package", "pkg5",
                        "portage", "pkgutil", "slackpkg", "swdepot", "zypper", "bundler", "pip",
                        "pear", "npm", "yarn", "gem", "easy_install", "bower", "package", "apk",
                        "openbsd_pkg", "pkgng", "sorcery", "xbps"]
    description = "package installs should use state=present with or without a version"

    if not errors:
        for task in tasks:
            if (task["action"]["__ansible_module__"] in package_managers
                    and task["action"].get("state") == "latest"):
                errors.append(Error(task["__line__"], description))

    return Result(candidate.path, errors)


def check_shell_instead_command(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    description = "shell should only be used when piping, redirecting or chaining commands"

    if not errors:
        for task in tasks:
            if task["action"]["__ansible_module__"] == "shell":
                if "cmd" in task["action"]:
                    cmd = task["action"].get("cmd", [])
                else:
                    cmd = " ".join(task["action"].get("__ansible_arguments__", []))

                unjinja = re.sub(r"\{\{[^\}]*\}\}", "JINJA_VAR", cmd)
                if not any([ch in unjinja for ch in "&|<>;$\n*[]{}?"]):
                    errors.append(Error(task["__line__"], description))

    return Result(candidate.path, errors)


def check_command_has_changes(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    commands = ["command", "shell", "raw"]
    description = "commands should either read information (and thus set changed_when) or not " \
                  "do something if it has already been done (using creates/removes) " \
                  "or only do it if another check has a particular result (when)"

    if not errors:
        for task in tasks:
            if task["action"]["__ansible_module__"] in commands:
                if ("changed_when" not in task and "when" not in task
                        and "when" not in task["__ansible_action_meta__"]
                        and "creates" not in task["action"]
                        and "removes" not in task["action"]):
                    errors.append(Error(task["__line__"], description))

    return Result(candidate.path, errors)


def check_empty_string_compare(candidate, settings):
    yamllines, errors = get_normalized_yaml(candidate, settings)
    description = "use `when: var` rather than `when: var != ""` (or " \
                  "conversely `when: not var` rather than `when: var == ""`)"

    if not errors:
        if isinstance(candidate, Template):
            matches = []
            jinja_string = re.compile("({{|{%)(.*?)(}}|%})")

            for i, line in yamllines:
                match = jinja_string.findall(line)
                if match:
                    for item in match:
                        matches.append((i, item[1]))

            yamllines = matches

        empty_string_compare = re.compile("[=!]= ?[\"'][\"']")

        for i, line in yamllines:
            if empty_string_compare.findall(line):
                errors.append(Error(i, description))

    return Result(candidate.path, errors)


def check_compare_to_literal_bool(candidate, settings):
    yamllines, errors = get_normalized_yaml(candidate, settings)
    description = "use `when: var` rather than `when: var == True` " \
                  "(or conversely `when: not var`)"

    if not errors:
        if isinstance(candidate, Template):
            matches = []
            jinja_string = re.compile("({{|{%)(.*?)(}}|%})")

            for i, line in yamllines:
                match = jinja_string.findall(line)
                if match:
                    for item in match:
                        matches.append((i, item[1]))

            yamllines = matches

        literal_bool_compare = re.compile("[=!]= ?(True|true|False|false)")

        for i, line in yamllines:
            if literal_bool_compare.findall(line):
                errors.append(Error(i, description))

    return Result(candidate.path, errors)


def check_delegate_to_localhost(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    description = "connection: local ensures that unexpected delegated_vars " \
                  "don't get set (e.g. {{ inventory_hostname }} " \
                  "used by vars_files)"

    if not errors:
        for task in tasks:
            if task.get("delegate_to") == "localhost":
                errors.append(Error(task["__line__"], description))

    return Result(candidate.path, errors)


def check_literal_bool_format(candidate, settings):
    yamllines, errors = get_normalized_yaml(candidate, settings)
    description = "literal bools should be written as 'True/False' or 'yes/no'"

    uppercase_bool = re.compile(r"([=!]=|:)\s*(true|false|TRUE|FALSE|Yes|No|YES|NO)\s*$")

    if not errors:
        for i, line in yamllines:
            if uppercase_bool.findall(line):
                errors.append(Error(i, description))

    return Result(candidate.path, errors)


def check_become_user(candidate, settings):
    tasks, errors = get_normalized_tasks(candidate, settings)
    description = "the task has 'become:' enabled but 'become_user:' is missing"
    true_value = [True, "true", "True", "TRUE", "yes", "Yes", "YES"]

    if not errors:
        gen = (task for task in tasks if "become" in task)
        for task in gen:
            if task["become"] in true_value and "become_user" not in task.keys():
                errors.append(Error(task["__line__"], description))

    return Result(candidate.path, errors)


def check_filter_separation(candidate, settings):
    yamllines, errors = get_normalized_yaml(candidate, settings)
    description = "no suitable numbers of spaces (required: 1)"

    matches = []
    braces = re.compile("{{(.*?)}}")
    filters = re.compile(r"(?<=\|)([\s]{2,}[^\s}]+|[^\s]+)|([^\s{]+[\s]{2,}|[^\s]+)(?=\|)")

    if not errors:
        for i, line in yamllines:
            match = braces.findall(line)
            if match:
                for item in match:
                    matches.append((i, item))

        for i, line in matches:
            if filters.findall(line):
                errors.append(Error(i, description))
    return Result(candidate.path, errors)



"""Global utils collection."""

from __future__ import print_function

import contextlib
import os
import re
import sys
from distutils.version import LooseVersion

import yaml

from ansiblelater import logger

try:
    import ConfigParser as configparser # noqa
except ImportError:
    import configparser # noqa


LOG = logger.get_logger(__name__)


def count_spaces(c_string):
    leading_spaces = 0
    trailing_spaces = 0

    for i, e in enumerate(c_string):
        if not e.isspace():
            break
        leading_spaces += 1

    for i, e in reversed(list(enumerate(c_string))):
        if not e.isspace():
            break
        trailing_spaces += 1

    return((leading_spaces, trailing_spaces))


def get_property(prop):
    currentdir = os.path.dirname(os.path.realpath(__file__))
    parentdir = os.path.dirname(currentdir)
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop),
        open(os.path.join(parentdir, "__init__.py")).read())
    return result.group(1)


def standards_latest(standards):
    return max([standard.version for standard in standards if standard.version] or ["0.1"],
               key=LooseVersion)


def lines_ranges(lines_spec):
    if not lines_spec:
        return None
    result = []
    for interval in lines_spec.split(","):
        (start, end) = interval.split("-")
        result.append(range(int(start), int(end) + 1))
    return result


def is_line_in_ranges(line, ranges):
    return not ranges or any([line in r for r in ranges])


def safe_load(string):
    """
    Parse the provided string returns a dict.

    :param string: A string to be parsed.
    :returns: dict

    """
    try:
        return yaml.safe_load(string) or {}
    except yaml.scanner.ScannerError as e:
        print(str(e))


@contextlib.contextmanager
def open_file(filename, mode="r"):
    """
    Open the provide file safely and returns a file type.

    :param filename: A string containing an absolute path to the file to open.
    :param mode: A string describing the way in which the file will be used.
    :returns: file type

    """
    with open(filename, mode) as stream:
        yield stream


def add_dict_branch(tree, vector, value):
    key = vector[0]
    tree[key] = value \
        if len(vector) == 1 \
        else add_dict_branch(tree[key] if key in tree else {},
                             vector[1:],
                             value)
    return tree


def sysexit(code=1):
    sys.exit(code)


def sysexit_with_message(msg, code=1):
    LOG.critical(msg)
    sysexit(code)



"""Standard definition."""


class Standard(object):
    """
    Standard definition for all defined rules.

    Later lookup the config file for a path to a rules directory
    or fallback to default `ansiblelater/data/*`.
    """

    def __init__(self, standard_dict):
        """
        Initialize a new standard object and returns None.

        :param standard_dict: Dictionary object containing all neseccary attributes

        """
        self.id = standard_dict.get("id", "")
        self.name = standard_dict.get("name")
        self.version = standard_dict.get("version")
        self.check = standard_dict.get("check")
        self.types = standard_dict.get("types")


    def __repr__(self): # noqa
        return "Standard: %s (version: %s, types: %s)" % (
               self.name, self.version, self.types)



"""Example standards definition."""

from ansiblelater.rules.ansiblefiles import check_become_user
from ansiblelater.rules.ansiblefiles import check_braces_spaces
from ansiblelater.rules.ansiblefiles import check_command_has_changes
from ansiblelater.rules.ansiblefiles import check_command_instead_of_module
from ansiblelater.rules.ansiblefiles import check_compare_to_literal_bool
from ansiblelater.rules.ansiblefiles import check_empty_string_compare
from ansiblelater.rules.ansiblefiles import check_filter_separation
from ansiblelater.rules.ansiblefiles import check_install_use_latest
from ansiblelater.rules.ansiblefiles import check_literal_bool_format
from ansiblelater.rules.ansiblefiles import check_name_format
from ansiblelater.rules.ansiblefiles import check_named_task
from ansiblelater.rules.ansiblefiles import check_shell_instead_command
from ansiblelater.rules.ansiblefiles import check_unique_named_task
from ansiblelater.rules.rolefiles import check_meta_main
from ansiblelater.rules.rolefiles import check_scm_in_src
from ansiblelater.rules.taskfiles import check_line_between_tasks
from ansiblelater.rules.yamlfiles import check_native_yaml
from ansiblelater.rules.yamlfiles import check_yaml_colons
from ansiblelater.rules.yamlfiles import check_yaml_document_end
from ansiblelater.rules.yamlfiles import check_yaml_document_start
from ansiblelater.rules.yamlfiles import check_yaml_empty_lines
from ansiblelater.rules.yamlfiles import check_yaml_file
from ansiblelater.rules.yamlfiles import check_yaml_has_content
from ansiblelater.rules.yamlfiles import check_yaml_hyphens
from ansiblelater.rules.yamlfiles import check_yaml_indent
from ansiblelater.standard import Standard

tasks_should_be_separated = Standard(dict(
    id="ANSIBLE0001",
    name="Single tasks should be separated by empty line",
    check=check_line_between_tasks,
    version="0.1",
    types=["playbook", "task", "handler"]
))

role_must_contain_meta_main = Standard(dict(
    id="ANSIBLE0002",
    name="Roles must contain suitable meta/main.yml",
    check=check_meta_main,
    version="0.1",
    types=["meta"]
))

tasks_are_uniquely_named = Standard(dict(
    id="ANSIBLE0003",
    name="Tasks and handlers must be uniquely named within a single file",
    check=check_unique_named_task,
    version="0.1",
    types=["playbook", "task", "handler"],
))

use_spaces_between_variable_braces = Standard(dict(
    id="ANSIBLE0004",
    name="YAML should use consistent number of spaces around variables",
    check=check_braces_spaces,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars", "meta"]
))

roles_scm_not_in_src = Standard(dict(
    id="ANSIBLE0005",
    name="Use scm key rather than src: scm+url",
    check=check_scm_in_src,
    version="0.1",
    types=["rolesfile"]
))

tasks_are_named = Standard(dict(
    id="ANSIBLE0006",
    name="Tasks and handlers must be named",
    check=check_named_task,
    version="0.1",
    types=["playbook", "task", "handler"],
))

tasks_names_are_formatted = Standard(dict(
    id="ANSIBLE0007",
    name="Name of tasks and handlers must be formatted",
    check=check_name_format,
    version="0.1",
    types=["playbook", "task", "handler"],
))

commands_should_not_be_used_in_place_of_modules = Standard(dict(
    id="ANSIBLE0008",
    name="Commands should not be used in place of modules",
    check=check_command_instead_of_module,
    version="0.1",
    types=["playbook", "task", "handler"]
))

package_installs_should_not_use_latest = Standard(dict(
    id="ANSIBLE0009",
    name="Package installs should use present, not latest",
    check=check_install_use_latest,
    types=["playbook", "task", "handler"]
))

use_shell_only_when_necessary = Standard(dict(
    id="ANSIBLE0010",
    name="Shell should only be used when essential",
    check=check_shell_instead_command,
    types=["playbook", "task", "handler"]
))

commands_should_be_idempotent = Standard(dict(
    id="ANSIBLE0011",
    name="Commands should be idempotent",
    check=check_command_has_changes,
    version="0.1",
    types=["playbook", "task"]
))

dont_compare_to_empty_string = Standard(dict(
    id="ANSIBLE0012",
    name="Don't compare to \"\" - use `when: var` or `when: not var`",
    check=check_empty_string_compare,
    version="0.1",
    types=["playbook", "task", "handler", "template"]
))

dont_compare_to_literal_bool = Standard(dict(
    id="ANSIBLE0013",
    name="Don't compare to True or False - use `when: var` or `when: not var`",
    check=check_compare_to_literal_bool,
    version="0.1",
    types=["playbook", "task", "handler", "template"]
))

literal_bool_should_be_formatted = Standard(dict(
    id="ANSIBLE0014",
    name="Literal bools should start with a capital letter",
    check=check_literal_bool_format,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars"]
))

use_become_with_become_user = Standard(dict(
    id="ANSIBLE0015",
    name="become should be combined with become_user",
    check=check_become_user,
    version="0.1",
    types=["playbook", "task", "handler"]
))

use_spaces_around_filters = Standard(dict(
    id="ANSIBLE0016",
    name="jinja2 filters should be separated with spaces",
    check=check_filter_separation,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars"]
))

files_should_not_contain_unnecessarily_empty_lines = Standard(dict(
    id="LINT0001",
    name="YAML should not contain unnecessarily empty lines",
    check=check_yaml_empty_lines,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars", "meta"]
))

files_should_be_indented = Standard(dict(
    id="LINT0002",
    name="YAML should be correctly indented",
    check=check_yaml_indent,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars", "meta"]
))

files_should_use_consistent_spaces_after_hyphens = Standard(dict(
    id="LINT0003",
    name="YAML should use consistent number of spaces after hyphens",
    check=check_yaml_hyphens,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars", "meta"]
))

files_should_contain_document_start_marker = Standard(dict(
    id="LINT0004",
    name="YAML should contain document start marker",
    check=check_yaml_document_start,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars", "meta"]
))

spaces_around_colons = Standard(dict(
    id="LINT0005",
    name="YAML should use consistent number of spaces around colons",
    check=check_yaml_colons,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars", "meta"]
))

rolesfile_should_be_in_yaml = Standard(dict(
    id="LINT0006",
    name="Roles file should be in yaml format",
    check=check_yaml_file,
    version="0.1",
    types=["rolesfile"]
))

files_should_not_be_purposeless = Standard(dict(
    id="LINT0007",
    name="Files should contain useful content",
    check=check_yaml_has_content,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars", "defaults", "meta"]
))

use_yaml_rather_than_key_value = Standard(dict(
    id="LINT0008",
    name="Use YAML format for tasks and handlers rather than key=value",
    check=check_native_yaml,
    version="0.1",
    types=["playbook", "task", "handler"]
))

files_should_contain_document_end_marker = Standard(dict(
    id="LINT0009",
    name="YAML should contain document end marker",
    check=check_yaml_document_end,
    version="0.1",
    types=["playbook", "task", "handler", "rolevars",
           "hostvars", "groupvars", "meta"]
))

ansible_min_version = "2.5"
ansible_later_min_version = "0.3.0"


standards = [
    # Ansible
    tasks_should_be_separated,
    role_must_contain_meta_main,
    tasks_are_uniquely_named,
    use_spaces_between_variable_braces,
    roles_scm_not_in_src,
    tasks_are_named,
    tasks_names_are_formatted,
    commands_should_not_be_used_in_place_of_modules,
    package_installs_should_not_use_latest,
    use_shell_only_when_necessary,
    commands_should_be_idempotent,
    dont_compare_to_empty_string,
    dont_compare_to_literal_bool,
    literal_bool_should_be_formatted,
    use_become_with_become_user,
    use_spaces_around_filters,
    # Lint
    files_should_not_contain_unnecessarily_empty_lines,
    files_should_be_indented,
    files_should_use_consistent_spaces_after_hyphens,
    files_should_contain_document_start_marker,
    spaces_around_colons,
    rolesfile_should_be_in_yaml,
    files_should_not_be_purposeless,
    use_yaml_rather_than_key_value,
    files_should_contain_document_end_marker,
]



"""Checks related to ansible task files."""

import re
from collections import defaultdict

from ansiblelater.command.candidates import Error
from ansiblelater.command.candidates import Result
from ansiblelater.utils.rulehelper import get_normalized_yaml


def check_line_between_tasks(candidate, settings):
    options = defaultdict(dict)
    options.update(remove_empty=False)
    options.update(remove_markers=False)

    lines, errors = get_normalized_yaml(candidate, settings, options)
    description = "missing task separation (required: 1 empty line)"

    task_regex = re.compile(r"-\sname:.*")
    prevline = "#file_start_marker"

    allowed_prevline = ["---", "tasks:", "pre_tasks:", "post_tasks:", "block:"]

    if not errors:
        for i, line in lines:
            match = task_regex.search(line)
            if match and prevline:
                if not any(item in prevline for item in allowed_prevline):
                    errors.append(Error(i, description))
            prevline = line.strip()

    return Result(candidate.path, errors)



"""Review candidates."""

import codecs
import copy
import os
import re
from distutils.version import LooseVersion

from six import iteritems

from ansiblelater import LOG
from ansiblelater import utils
from ansiblelater.logger import flag_extra

try:
    # Ansible 2.4 import of module loader
    from ansible.plugins.loader import module_loader
except ImportError:
    try:
        from ansible.plugins import module_loader
    except ImportError:
        from ansible.utils import module_finder as module_loader


class Candidate(object):
    """
    Meta object for all files which later has to process.

    Each file passed to later will be classified by type and
    bundled with necessary meta informations for rule processing.
    """

    def __init__(self, filename, settings={}, standards=[]):
        self.path = filename
        self.binary = False
        self.vault = False
        self.filetype = type(self).__name__.lower()
        self.expected_version = True
        self.standards = self._get_standards(settings, standards)

        try:
            with codecs.open(filename, mode="rb", encoding="utf-8") as f:
                if f.readline().startswith("$ANSIBLE_VAULT"):
                    self.vault = True
        except UnicodeDecodeError:
            self.binary = True

        self.version = self._get_version(settings)

    def _get_version(self, settings):
        path = self.path
        version = None

        if not self.binary:
            if isinstance(self, RoleFile):
                parentdir = os.path.dirname(os.path.abspath(self.path))
                while parentdir != os.path.dirname(parentdir):
                    meta_file = os.path.join(parentdir, "meta", "main.yml")
                    if os.path.exists(meta_file):
                        path = meta_file
                        break
                    parentdir = os.path.dirname(parentdir)

            version_re = re.compile(r"^# Standards:\s*([\d.]+)")

            with codecs.open(path, mode="rb", encoding="utf-8") as f:
                for line in f:
                    match = version_re.match(line)
                    if match:
                        version = match.group(1)

        if not version:
            version = utils.standards_latest(self.standards)
            if self.expected_version:
                if isinstance(self, RoleFile):
                    LOG.warning(
                        "%s %s is in a role that contains a meta/main.yml without a declared "
                        "standards version. "
                        "Using latest standards version %s" %
                        (type(self).__name__, self.path, version))
                else:
                    LOG.warning(
                        "%s %s does not present standards version. "
                        "Using latest standards version %s" %
                        (type(self).__name__, self.path, version))
        else:
            LOG.info("%s %s declares standards version %s" %
                     (type(self).__name__, self.path, version))

        return version

    def _get_standards(self, settings, standards):
        target_standards = []
        includes = settings.config["rules"]["filter"]
        excludes = settings.config["rules"]["exclude_filter"]

        if len(includes) == 0:
            includes = [s.id for s in standards]

        for standard in standards:
            if standard.id in includes and standard.id not in excludes:
                target_standards.append(standard)

        return target_standards

    def review(self, settings, lines=None):
        errors = 0

        for standard in self.standards:
            if type(self).__name__.lower() not in standard.types:
                continue

            result = standard.check(self, settings.config)

            if not result:
                utils.sysexit_with_message("Standard '{}' returns an empty result object.".format(
                    standard.check.__name__))

            labels = {"tag": "review", "standard": standard.name, "file": self.path, "passed": True}

            if standard.id and standard.id.strip():
                labels["id"] = standard.id

            for err in [err for err in result.errors
                        if not err.lineno or utils.is_line_in_ranges(err.lineno, utils.lines_ranges(lines))]: # noqa
                err_labels = copy.copy(labels)
                err_labels["passed"] = False
                if isinstance(err, Error):
                    err_labels.update(err.to_dict())

                if not standard.version:
                    LOG.warning("{id}Best practice '{name}' not met:\n{path}:{error}".format(
                        id=self._format_id(standard.id),
                        name=standard.name,
                        path=self.path,
                        error=err), extra=flag_extra(err_labels))
                elif LooseVersion(standard.version) > LooseVersion(self.version):
                    LOG.warning("{id}Future standard '{name}' not met:\n{path}:{error}".format(
                        id=self._format_id(standard.id),
                        name=standard.name,
                        path=self.path,
                        error=err), extra=flag_extra(err_labels))
                else:
                    LOG.error("{id}Standard '{name}' not met:\n{path}:{error}".format(
                        id=self._format_id(standard.id),
                        name=standard.name,
                        path=self.path,
                        error=err), extra=flag_extra(err_labels))
                    errors = errors + 1

        return errors

    def _format_id(self, standard_id):
        if standard_id and standard_id.strip():
            standard_id = "[{id}] ".format(id=standard_id.strip())

        return standard_id

    def __repr__(self): # noqa
        return "%s (%s)" % (type(self).__name__, self.path)

    def __getitem__(self, item): # noqa
        return self.__dict__.get(item)


class RoleFile(Candidate):
    def __init__(self, filename, settings={}, standards=[]):
        super(RoleFile, self).__init__(filename, settings, standards)

        parentdir = os.path.dirname(os.path.abspath(filename))
        while parentdir != os.path.dirname(parentdir):
            role_modules = os.path.join(parentdir, "library")
            if os.path.exists(role_modules):
                module_loader.add_directory(role_modules)
                break
            parentdir = os.path.dirname(parentdir)


class Playbook(Candidate):
    pass


class Task(RoleFile):
    def __init__(self, filename, settings={}, standards=[]):
        super(Task, self).__init__(filename, settings, standards)
        self.filetype = "tasks"


class Handler(RoleFile):
    def __init__(self, filename, settings={}, standards=[]):
        super(Handler, self).__init__(filename, settings, standards)
        self.filetype = "handlers"


class Vars(Candidate):
    pass


class Unversioned(Candidate):
    def __init__(self, filename, settings={}, standards=[]):
        super(Unversioned, self).__init__(filename, settings, standards)
        self.expected_version = False


class InventoryVars(Unversioned):
    pass


class HostVars(InventoryVars):
    pass


class GroupVars(InventoryVars):
    pass


class RoleVars(RoleFile):
    pass


class Meta(RoleFile):
    pass


class Inventory(Unversioned):
    pass


class Code(Unversioned):
    pass


class Template(RoleFile):
    pass


class Doc(Unversioned):
    pass


class Makefile(Unversioned):
    pass


class File(RoleFile):
    pass


class Rolesfile(Unversioned):
    pass


class Error(object):
    """Default error object created if a rule failed."""

    def __init__(self, lineno, message, error_type=None, **kwargs):
        """
        Initialize a new error object and returns None.

        :param lineno: Line number where the error from de rule occures
        :param message: Detailed error description provided by the rule

        """
        self.lineno = lineno
        self.message = message
        self.kwargs = kwargs
        for (key, value) in iteritems(kwargs):
            setattr(self, key, value)

    def __repr__(self): # noqa
        if self.lineno:
            return "%s: %s" % (self.lineno, self.message)
        else:
            return " %s" % (self.message)

    def to_dict(self):
        result = dict(lineno=self.lineno, message=self.message)
        for (key, value) in iteritems(self.kwargs):
            result[key] = value
        return result


class Result(object):
    def __init__(self, candidate, errors=None):
        self.candidate = candidate
        self.errors = errors or []

    def message(self):
        return "\n".join(["{0}:{1}".format(self.candidate, error)
                          for error in self.errors])


def classify(filename, settings={}, standards=[]):
    parentdir = os.path.basename(os.path.dirname(filename))
    basename = os.path.basename(filename)

    if parentdir in ["tasks"]:
        return Task(filename, settings, standards)
    if parentdir in ["handlers"]:
        return Handler(filename, settings, standards)
    if parentdir in ["vars", "defaults"]:
        return RoleVars(filename, settings, standards)
    if "group_vars" in filename.split(os.sep):
        return GroupVars(filename, settings, standards)
    if "host_vars" in filename.split(os.sep):
        return HostVars(filename, settings, standards)
    if parentdir in ["meta"]:
        return Meta(filename, settings, standards)
    if parentdir in ["library", "lookup_plugins", "callback_plugins",
                     "filter_plugins"] or filename.endswith(".py"):
        return Code(filename, settings, standards)
    if "inventory" == basename or "hosts" == basename or parentdir in ["inventories"]:
        return Inventory(filename, settings, standards)
    if "rolesfile" in basename or "requirements" in basename:
        return Rolesfile(filename, settings, standards)
    if "Makefile" in basename:
        return Makefile(filename, settings, standards)
    if "templates" in filename.split(os.sep) or basename.endswith(".j2"):
        return Template(filename, settings, standards)
    if "files" in filename.split(os.sep):
        return File(filename, settings, standards)
    if basename.endswith(".yml") or basename.endswith(".yaml"):
        return Playbook(filename, settings, standards)
    if "README" in basename:
        return Doc(filename, settings, standards)
    return None
